#include <limits.h>
#include <stdio.h>
#include <string.h>
#include "conf_data.h"
#include "parse_conf.h"

void set_error(/*@out@*/ struct service_config *sc, int lineno,
		const char *message) {
	(void)snprintf(sc->parse_error, sizeof(sc->parse_error), "line %d: %s",
			lineno, message);
}

int parse_bool(const char *value) {
	if(0 == strcmp(value, "0"))
		return 0;
	if(0 == strcmp(value, "no"))
		return 0;
	if(0 == strcmp(value, "false"))
		return 0;
	if(0 == strcmp(value, "off"))
		return 0;
	if(0 == strcmp(value, "1"))
		return 1;
	if(0 == strcmp(value, "yes"))
		return 1;
	if(0 == strcmp(value, "true"))
		return 1;
	if(0 == strcmp(value, "on"))
		return 1;
	fprintf(stderr, "value:: |%s|\n", value);
	return 2;
}

int parse_assignment(int lineno, const char *section, const char *key,
		const char *value, /*@out@*/ struct service_config *sc) {
	if(0 == strcmp(section, "Install"))
		/* We don't care about install stuff. */
		return 0;
	if(0 == strcmp(section, "Unit")) {
		if(0 == strcmp(key, "After")) {
			if(sc->after[0] != '\0') {
				set_error(sc, lineno,
						"multiple After not supported");
				return -1;
			}
			if(value[0] == '\0') {
				set_error(sc, lineno,
						"empty After not supported");
				return -1;
			}
			strncpy(sc->after, value, sizeof(sc->after));
			return 0;
		}
		if(0 == strcmp(key, "Before")) {
			if(sc->before[0] != '\0') {
				set_error(sc, lineno,
						"multiple Before not supported");
				return -1;
			}
			if(value[0] == '\0') {
				set_error(sc, lineno,
						"empty Before not supported");
				return -1;
			}
			strncpy(sc->before, value, sizeof(sc->before));
			return 0;
		}
		if(0 == strcmp(key, "DefaultDependencies")) {
			switch(parse_bool(value)) {
				case 0:
					sc->defaultdependencies = DD_EXCLUDE;
					return 0;
				case 1:
					sc->defaultdependencies = DD_INCLUDE;
					return 0;
				default:
					set_error(sc, lineno,
							"invalid boolean");
					return -1;
				}
		}
		if(0 == strcmp(key, "Description")) {
			strncpy(sc->description, value,
					sizeof(sc->description));
			return 0;
		}
	} else if(0 == strcmp(section, "Service")) {
		if(0 == strcmp(key, "ExecStart")) {
			strncpy(sc->exec_start, value, sizeof(sc->exec_start));
			return 0;
		} else if(0 == strcmp(key, "Type")) {
			if(0 == strcmp(value, "simple")) {
				sc->type = ST_SIMPLE;
				return 0;
			} if(0 == strcmp(value, "oneshot")) {
				sc->type = ST_ONESHOT;
				return 0;
			} else {
				set_error(sc, lineno, "invalid service type");
				return -1;
			}
		}
	}
	(void)snprintf(sc->parse_error, sizeof(sc->parse_error),
			"line %d: section %s key %s unknown", lineno, section,
			key);
	return -1;
}

int parse_stream(FILE *f, /*@out@*/ struct service_config *sc) {
	char line[LINE_MAX], section[LINE_MAX], *p;
	int lineno=0;
	while(NULL != fgets(line, sizeof(line), f)) {
		++lineno;
		p = line + strlen(line) - 1;
		if(*p != '\n') {
			set_error(sc, lineno, "line too long");
			return -1;
		}
		*p = '\0';
		if(line[0] == '\0' || line[0] == '#' || line[0] == ';')
			continue;
		if(line[0] == '[') {
			if(NULL == (p = strchr(line, ']'))) {
				set_error(sc, lineno, "unclosed section");
				return -1;
			}
			if(p[1] != '\0') {
				set_error(sc, lineno, "garbage after section");
				return -1;
			}
			*p = '\0';
			strcpy(section, line + 1);
			if(0 != strcmp(section, "Unit") &&
					0 != strcmp(section, "Service") &&
					0 != strcmp(section, "Install")) {
				set_error(sc, lineno, "invalid section name");
				return -1;
			}
			continue;
		}
		if(NULL == (p = strchr(line, '='))) {
			set_error(sc, lineno, "non-assignment");
			return -1;
		}
		*p++ = '\0';
		if(p[strlen(p) - 1] == '\\') {
			set_error(sc, lineno, "continuation");
			return -1;
		}
		if(parse_assignment(lineno, section, line, p, sc) < 0)
			return -1;
	}
	if(ferror(f) == 0)
		return 0;
	return -1;
}

int parse_file(const char *filename, /*@out@*/ struct service_config *sc) {
	FILE *f;
	/* TODO: support service masking (size == 0) */
	if(NULL == (f = fopen(filename, "r")))
		return -1;
	if(parse_stream(f, sc) < 0) {
		(void)fclose(f);
		return -1;
	}
	if(fclose(f) < 0)
		return -1;
	return 0;
}
